#!/usr/bin/env python
#
# laika prints a summary of the Launchpad bugs you touched this week
#
# man's best friend guides you through Launchpad
# 'Laika died within hours after launch' => an aptly named tool ;)
# http://en.wikipedia.org/wiki/Laika
#
# Copyright 2010 Alex Chiang <achiang@canonical.com>
#
# This program is distributed under the terms of the
# GNU General Public License version 2.

import ConfigParser
import datetime
from optparse import OptionParser
import os
import re
import sys

from launchpadlib.launchpad import Launchpad

UTCNOW = datetime.datetime.utcnow()


class Report(object):
    '''An activity report for a specified Launchpad user.

    In order of decreasing importance, search and display activity on
    bugs assigned to me, then on bugs that I actually commented upon,
    and finally bugs that I just opened, but did nothing further with.

    Avoids duplication of activity. So if the status was changed on
    a bug assigned to you and you also commented on it, only report the
    status change.
    '''

    bugs = {}

    def __init__(self, user, window, ppas):
        cachedir = os.path.expanduser("~/.launchpadlib/cache")
        self.launchpad = Launchpad.login_with('laika', 'production', cachedir)

        self.user = self.launchpad.people[user]
        self.window = window
        self.ppas = ppas

    def print_header(self, header):
        print "==", header, "=="

    def in_window(self, date):
        '''Timezones do not exist, all datetime objects have to be naive.

        Time zone aware means broken.
        http://www.enricozini.org/2009/debian/using-python-datetime/
        '''
        win = datetime.timedelta(self.window)
        date = date.replace(tzinfo=None)
        delta = UTCNOW - date
        return delta <= win

    def print_bugid(self, task):
        '''Using this interface adds the bug to global bug list.'''
        ago = ""
        self.bugs[task.bug.id] = 1
        delta = UTCNOW - task.bug.date_last_updated.replace(tzinfo=None)
        if delta.days > 0:
            ago = "%d day%s" % (delta.days, "s" if delta.days > 1 else "")

        hours = delta.seconds / 3600
        if hours > 0:
            ago += ", " if ago else ""
            ago += "%d hour%s" % (hours, "s" if hours > 1 else "")

        minutes = (delta.seconds - (hours * 3600)) / 60
        if minutes > 0:
            ago += ", " if ago else ""
            ago += "%d minute%s" % (minutes, "s" if minutes > 1 else "")

        print task.title
        print "https://launchpad.net/bugs/" + str(task.bug.id)
        print "last updated", ago, "ago"

    def print_assignments(self):
        statuses = ['closed', 'fix_released', 'fix_committed',
                'in_progress', 'triaged', 'confirmed', 'created']

        tasks = self.user.searchTasks(assignee=self.user)
        self.print_header("Assigned Bugs")

        for t in tasks:
            if self.bugs.has_key(t.bug.id):
                continue
            updates = []
            for s in statuses:
                attr = 'date_' + s
                date = getattr(t, attr)
                if not date:
                    continue
                if self.in_window(date):
                    updates.append(
                        "\t" + s + ": " + re.sub("\s.*$", "", str(date)))

            if updates:
                self.print_bugid(t)
                for u in updates:
                    print u
                print
        print

    def print_comments(self):
        tasks = self.user.searchTasks(bug_commenter=self.user)
        self.print_header("Commented Bugs")
        for t in tasks:
            if self.bugs.has_key(t.bug.id):
                continue
            for m in t.bug.messages:
                if m.owner_link != self.user.self_link:
                    continue
                if self.in_window(m.date_created):
                    self.print_bugid(t)
                    print
                    break
        print

    def print_reported(self):
        tasks = self.user.searchTasks(bug_reporter=self.user)
        self.print_header("Reported Bugs")
        for t in tasks:
            if self.bugs.has_key(t.bug.id):
                continue
            if self.in_window(t.bug.date_created):
                self.print_bugid(t)
                print

    def print_ppa_activity(self):
        if self.ppas == '':
            return

        # Handle command line case, where user passes in comma-separated list
        self.ppas = re.sub(",", "\n", self.ppas)

        # Now each PPA is separated by a newline, which is how we might
        # expect them if they're specified in the config file.
        ppas = [p for p in self.ppas.split("\n") if p != '']

        self.print_header("PPA Activity")
        for ppa in ppas:
            # A PPA seems to always belong to a person (or a team, which
            # is a person in LP) so assuming "person/PPA" sounds reasonable.
            person, archive = ppa.split('/')
            owner = self.launchpad.people[person]
            archive = owner.getPPAByName(name=archive)
            sources = archive.getPublishedSources()

            print ppa
            for s in sources:
                if s.package_signer != self.user:
                    continue

                # There might be yet-unpublished packages in the PPA, which
                # would then have NoneType for date_published. Don't check
                # in_window in that case.
                if s.date_published:
                    if self.in_window(s.date_published):
                        print "\t", s.display_name
            print
        print

    def render(self):
        self.print_assignments()
        self.print_comments()
        self.print_reported()
        self.print_ppa_activity()

def get_config(opts):
    config = {}
    FORMAT = "1.0"
    laika_config = ConfigParser.ConfigParser()
    laika_config.read(os.path.expanduser("~/.laikarc"))

    if laika_config.has_section(FORMAT) == False:
        print "Error: unsupported configuration file format"
        sys.exit()

    def default_user():
        return os.getenv("USER")
    def default_window():
        return 8
    def default_ppas():
        return ''

    for opt in opts:
        try:
            config[opt] = laika_config.get(FORMAT, opt)
        except ConfigParser.NoOptionError:
            config[opt] = locals()["default_%s" % opt]()

    return config

def main():
    opts = ['user', 'window', 'ppas']
    config = get_config(opts)

    parser = OptionParser()
    parser.add_option('-u', '--user', dest='user',
        default=config['user'],
        help='Specify the Launchpad user id.  '
             'Defaults to session username.')
    parser.add_option('-w', '--window', dest='window', type='int',
        default=config['window'],
        help='Number of days of past activity to look for.  '
             'Defaults to 8 days')
    parser.add_option('--ppas', dest='ppas',
        default=config['ppas'],
        help='Search for activity in specified PPAs')

    options, arguments = parser.parse_args()

    report = Report(options.user, options.window, options.ppas)
    report.render()

if __name__ == "__main__":
    main()
